# SPDX-FileCopyrightText: 2024-present Canterbury Air Patrol Inc <github@canterburyairpatrol.org>
#
# SPDX-License-Identifier: MIT
"""
Search Management Map - Types
"""

MIN_LATITUDE = -90.0
MAX_LATITUDE = 90.0
MIN_LONGITUDE = -180.0
MAX_LONGITUDE = 180.0


class LatitudeError(ValueError):
    """
    Class for errors with latitudes
    """

    def __init__(self) -> None:
        super().__init__(f"Latitude out of range ({MIN_LATITUDE}, {MAX_LATITUDE} degrees)")


class LongitudeError(ValueError):
    """
    Class for errors with longitude
    """

    def __init__(self) -> None:
        super().__init__(f"Longitude out of range ({MIN_LONGITUDE}, {MAX_LONGITUDE} degrees)")


class SMMPoint:
    # pylint: disable=R0903
    """
    Latitude/Longitude combination
    """

    def __init__(self, latitude: float, longitude: float) -> None:
        self._lat = None
        self._lng = None
        self.lat = latitude
        self.lng = longitude

    def __set_lat(self, lat: float) -> None:
        if lat < MIN_LATITUDE or lat > MAX_LATITUDE:
            raise LatitudeError
        self._lat = lat

    def __get_lat(self) -> float:
        """
        Get the latitude
        """
        return self._lat

    def __set_lng(self, lng: float) -> None:
        if lng < MIN_LONGITUDE or lng > MAX_LONGITUDE:
            raise LongitudeError
        self._lng = lng

    def __get_lng(self) -> float:
        return self._lng

    lat = property(__get_lat, __set_lat)
    latitude = property(__get_lat, __set_lat)
    lng = property(__get_lng, __set_lng)
    longitude = property(__get_lng, __set_lng)



# SPDX-FileCopyrightText: 2024-present Canterbury Air Patrol Inc <github@canterburyairpatrol.org>
#
# SPDX-License-Identifier: MIT
"""
Search Management Map Connection Handlers
"""

from __future__ import annotations

import requests

from smm_client.assets import SMMAsset, SMMAssetStatusValue, SMMAssetType
from smm_client.missions import SMMMission, SMMMissionAssetStatusValue
from smm_client.organizations import SMMOrganization


# pylint: disable = R0903
class SMMUser:
    """
    User in Search Management Map
    """

    def __init__(self, user_id: int, username: str) -> None:
        self.id = user_id
        self.username = username


class SMMConnection:
    # pylint: disable=R0904
    """
    Create a connection to the Search Management Map Server
    """

    def __init__(self, url: str, username: str, password: str) -> None:
        self.base_url = url
        self.username = username
        self.password = password
        self.session = requests.Session()
        self.login()

    def get(self, path=None) -> requests.Response:
        """
        GET path on this connection
        """
        url = f"{self.base_url}/{path}" if path else self.base_url
        return self.session.get(url)

    def get_json(self, path: str):
        """
        GET json from path on this connection
        """
        url = f"{self.base_url}/{path}"
        return self.session.get(url, headers={"Accept": "application/json"}).json()

    def post(self, path: str, data=None) -> requests.Response:
        """
        POST data to path on this connection
        """
        url = f"{self.base_url}/{path}"
        return self.session.post(url, data=data, headers={"X-CSRFToken": self.session.cookies["csrftoken"]})

    def delete(self, path: str) -> requests.Response:
        """
        DELETE path on this connection
        """
        url = f"{self.base_url}/{path}"
        return self.session.delete(url, headers={"X-CSRFToken": self.session.cookies["csrftoken"]})

    def login(self) -> None:
        """
        Login to the server
        """
        self.get()
        self.post("/accounts/login/", data={"username": self.username, "password": self.password})

    def get_assets(self) -> list[SMMAsset]:
        """
        Get all the assests associated with the logged in user
        """
        assets_json = self.get_json("/assets/")["assets"]
        return [SMMAsset(self, asset_json["id"], asset_json["name"]) for asset_json in assets_json]

    def get_missions(self, only: str = "all") -> list[SMMMission]:
        """
        Get all the missions the logged in user is a member of
        """
        missions_json = self.get_json(f"/mission/list/?only={only}")["missions"]
        return [SMMMission(self, mission_json["id"], mission_json["name"]) for mission_json in missions_json]

    def get_asset_types(self) -> list[SMMAssetType]:
        """
        Get all asset types
        """
        asset_types_json = self.get_json("/assets/assettypes/")["asset_types"]
        return [
            SMMAssetType(self, asset_type_json["id"], asset_type_json["name"]) for asset_type_json in asset_types_json
        ]

    def create_asset_status_value(self, name: str, description: str, *, inop: bool) -> SMMAssetStatusValue:
        """
        Add an asset status value
        """
        result = self.post(
            "/admin/assets/assetstatusvalue/add/",
            {"name": name, "description": description, "inop": inop, "_continue": "Save+and+continue+editing"},
        )
        url_parts = result.url.split("/")
        return SMMAssetStatusValue(url_parts[-3], name, description, inop=inop)

    def get_asset_status_values(self) -> list[SMMAssetStatusValue]:
        """
        Get all the asset status values
        """
        asset_status_values_json = self.get_json("/assets/status/values/")["values"]
        return [
            SMMAssetStatusValue(
                asset_status_value["id"],
                asset_status_value["name"],
                asset_status_value["description"],
                inop=asset_status_value["inop"],
            )
            for asset_status_value in asset_status_values_json
        ]

    def get_or_create_asset_status_value(self, name: str, description: str, *, inop: bool) -> SMMAssetStatusValue:
        """
        Get the asset status value that matches this name
        Otherwise create it
        """
        status_values = self.get_asset_status_values()
        for sv in status_values:
            if sv.name == name:
                return sv
        return self.create_asset_status_value(name, description, inop=inop)

    def create_mission_asset_status_value(self, name: str, description: str) -> SMMMissionAssetStatusValue:
        """
        Add a mission asset status value
        """
        result = self.post(
            "/admin/mission/missionassetstatusvalue/add/",
            {"name": name, "description": description, "_continue": "Save+and+continue+editing"},
        )
        url_parts = result.url.split("/")
        return SMMMissionAssetStatusValue(url_parts[-3], name, description)

    def get_mission_asset_status_values(self) -> list[SMMMissionAssetStatusValue]:
        """
        Get all the mission asset status values
        """
        mission_asset_status_values_json = self.get_json("/mission/asset/status/values/")["values"]
        return [
            SMMMissionAssetStatusValue(
                asset_status_value["id"], asset_status_value["name"], asset_status_value["description"]
            )
            for asset_status_value in mission_asset_status_values_json
        ]

    def get_or_create_mission_asset_status_value(self, name: str, description: str) -> SMMMissionAssetStatusValue:
        """
        Get the mission asset status value that matches this name
        Otherwise create it
        """
        status_values = self.get_mission_asset_status_values()
        for sv in status_values:
            if sv.name == name:
                return sv
        return self.create_mission_asset_status_value(name, description)

    def get_organizations(self, *, all_orgs=False) -> list[SMMOrganization]:
        """
        Get all Organizations
        """
        url = "/organization/" if all_orgs else "/organization/?only=mine"
        organizations_json = self.get_json(url)["organizations"]
        return [
            SMMOrganization(self, organization_json["id"], organization_json["name"])
            for organization_json in organizations_json
        ]

    def create_user(self, username: str, password: str) -> SMMUser:
        """
        Add a new user to this server
        """
        result = self.post(
            "/admin/auth/user/add/",
            {"username": username, "password1": password, "password2": password, "_save": "Save"},
        )
        url_parts = result.url.split("/")
        return SMMUser(url_parts[-3], username)

    def create_asset_type(self, asset_type: str, description: str) -> SMMAssetType:
        """
        Create a new asset type
        """
        result = self.post(
            "/admin/assets/assettype/add/",
            {"name": asset_type, "description": description, "_continue": "Save+and+continue+editing"},
        )
        url_parts = result.url.split("/")
        return SMMAssetType(self, url_parts[-3], asset_type)

    def get_or_create_asset_type(self, asset_type: str, description: str) -> SMMAssetType:
        """
        Get the asset type that matches this asset type
        Otherwise create it
        """
        asset_types = self.get_asset_types()
        for at in asset_types:
            if at.name == asset_type:
                return at
        return self.create_asset_type(asset_type, description)

    def create_asset(self, user: SMMUser, asset: str, asset_type: SMMAssetType) -> SMMAsset:
        """
        Create a new asset
        """
        result = self.post(
            "/admin/assets/asset/add/",
            {"name": asset, "owner": user.id, "asset_type": asset_type.id, "_continue": "Save+and+continue+editing"},
        )
        url_parts = result.url.split("/")
        return SMMAsset(self, url_parts[-3], asset)

    def create_mission(self, name: str, description: str) -> SMMMission | None:
        """
        Create a new mission
        """
        res = self.post("/mission/new/", {"mission_name": name, "mission_description": description})
        if res.status_code == requests.codes["ok"]:
            return SMMMission(self, res.url.split("/")[-3], name)
        return None

    def create_organization(self, name: str) -> SMMOrganization:
        """
        Create a new organization
        """
        org_json = self.post("/organization/", {"name": name}).json()
        return SMMOrganization(self, org_json["id"], org_json["name"])

    def get_or_create_organization(self, name: str) -> SMMOrganization:
        """
        Get the organization that matches name
        Will be created if it doesn't already exist
        """
        organizations = self.get_organizations()
        for org in organizations:
            if org.name == name:
                return org
        return self.create_organization(name)



# SPDX-FileCopyrightText: 2024-present Canterbury Air Patrol Inc <github@canterburyairpatrol.org>
#
# SPDX-License-Identifier: MIT
"""
Search Management Map - Search
"""

from __future__ import annotations

from json import JSONDecodeError
from typing import TYPE_CHECKING

from smm_client.types import SMMPoint

if TYPE_CHECKING:
    from smm_client.assets import SMMAsset
    from smm_client.connection import SMMConnection


class SMMSearchData:
    # pylint: disable=R0903
    """
    Data for a specific search
    """

    def __init__(self, geojson):
        self.id = geojson["id"]
        self.properties = geojson["properties"]
        self.coords = [SMMPoint(p[1], p[0]) for p in geojson["geometry"]["coordinates"]]


class SMMSearch:
    """
    Base class for searches
    """

    def __init__(self, connection: SMMConnection, search_id: int):
        self.connection = connection
        self.id = search_id

    def __url_component(self, page: str | None) -> str:
        return f"/search/{self.id}/{page}" if page else f"/search/{self.id}/"

    def get_data(self) -> SMMSearchData | None:
        """
        Get the data for this search
        """
        res = self.connection.get_json(self.__url_component(None))
        try:
            return SMMSearchData(res["features"][0])
        except KeyError:
            return None

    def queue(self, asset: SMMAsset | None) -> bool:
        """
        Queue this search for a specific asset, or just for the asset type
        """
        if asset is not None:
            res = self.connection.post(self.__url_component("queue/"), data={"asset": asset.id})
        else:
            res = self.connection.post(self.__url_component("queue/"))
        return res.text == "Success"

    def begin(self, asset: SMMAsset) -> SMMSearchData | None:
        """
        Begin this search with asset
        """
        res = self.connection.post(self.__url_component("begin/"), data={"asset_id": asset.id})
        try:
            return SMMSearchData(res.json()["features"][0])
        except JSONDecodeError:
            return None

    def finished(self, asset: SMMAsset) -> bool:
        """
        Mark this search as finished/completed by asset
        """
        res = self.connection.post(self.__url_component("finished/"), data={"asset_id": asset.id})
        return res.text == "Completed"



# SPDX-FileCopyrightText: 2024-present Canterbury Air Patrol Inc <github@canterburyairpatrol.org>
#
# SPDX-License-Identifier: MIT
"""
Search Management Map - Missions
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import requests

from smm_client.geometry import SMMLine, SMMPoi, SMMPolygon
from smm_client.organizations import SMMOrganization

if TYPE_CHECKING:
    from smm_client.assets import SMMAsset, SMMUser
    from smm_client.connection import SMMConnection
    from smm_client.types import SMMPoint


class SMMMissionAssetStatusValue:
    # pylint: disable=R0903
    """
    Search Management Map - Mission Asset Status Value
    """

    def __init__(self, value_id: int, name: str, description: str) -> None:
        self.id = value_id
        self.name = name
        self.description = description


class SMMMissionOrganization:
    """
    Search Management Map - Organization membership of a Mission
    """

    def __init__(self, mission: SMMMission, organization: SMMOrganization) -> None:
        self.mission = mission
        self.organization = organization

    def set_can_add_organizations(self, *, value: bool) -> bool:
        """
        Set whether this organization can add organizations or not
        """
        response = self.mission.post(f"organizations/{self.organization.id}/", {"add_organization": value})
        return response.status_code == requests.codes["ok"]

    def set_can_add_users(self, *, value: bool) -> bool:
        """
        Set whether this organization can add members or not
        """
        response = self.mission.post(f"organizations/{self.organization.id}/", {"add_user": value})
        return response.status_code == requests.codes["ok"]


class SMMMissionMember:
    """
    Search Management Map - User membership of a Mission
    """

    def __init__(self, mission: SMMMission, user: SMMUser) -> None:
        self.mission = mission
        self.user = user

    def set_is_admin(self, *, value: bool) -> bool:
        """
        Set whether this user is an admin or not
        Admins have all other permissions as well
        """
        response = self.mission.post(f"users/{self.user.id}/", {"admin": value})
        return response.status_code == requests.codes["ok"]

    def set_can_add_organizations(self, *, value: bool) -> bool:
        """
        Set whether this organization can add organizations or not
        """
        response = self.mission.post(f"users/{self.user.id}/", {"add_organization": value})
        return response.status_code == requests.codes["ok"]

    def set_can_add_users(self, *, value: bool) -> bool:
        """
        Set whether this organization can add members or not
        """
        response = self.mission.post(f"users/{self.user.id}/", {"add_user": value})
        return response.status_code == requests.codes["ok"]


class SMMMission:
    """
    Search Management Map - Mission
    """

    def __init__(self, connection: SMMConnection, mission_id: int, name: str) -> None:
        self.connection = connection
        self.id = mission_id
        self.name = name

    def __str__(self) -> str:
        return f"{self.name} ({self.id})"

    def __url_component(self, page: str) -> str:
        return f"/mission/{self.id}/{page}"

    def post(self, page: str, data: object):
        """
        Post data to a specific url in this mission
        """
        return self.connection.post(self.__url_component(page), data)

    def get_json(self, page: str):
        """
        Get data from a specific url in this mission
        """
        return self.connection.get_json(self.__url_component(page))

    def add_member(self, user: SMMUser) -> SMMMissionMember:
        """
        Add a member to this mission
        """
        self.post("users/add/", data={"user": user.username})
        return SMMMissionMember(self, user)

    def add_organization(self, organization: SMMOrganization) -> SMMMissionOrganization:
        """
        Add an organization to this mission
        """
        self.post("organizations/", data={"organization": organization.id})
        return SMMMissionOrganization(self, organization)

    def get_organizations(self) -> list[SMMMissionOrganization]:
        """
        Get all the current organizations in this mission
        """
        organizations_json = self.get_json("organizations/")
        return [
            SMMMissionOrganization(
                self,
                SMMOrganization(
                    self.connection, organization["organization"]["id"], organization["organization"]["name"]
                ),
            )
            for organization in organizations_json["organizations"]
        ]

    def add_asset(self, asset: SMMAsset) -> None:
        """
        Add an asset to this mission
        """
        self.post("assets/", data={"asset": asset.id})

    def remove_asset(self, asset: SMMAsset) -> None:
        """
        Remove an asset from this mission
        """
        self.connection.get(self.__url_component(f"assets/{asset.id}/remove/"))

    def set_asset_command(self, asset: SMMAsset, command: str, reason: str, point: SMMPoint | None = None) -> None:
        """
        Set the command for a specific asset
        """
        data = {
            "asset": asset.id,
            "command": command,
            "reason": reason,
        }
        if point is not None:
            data["latitude"] = point.latitude
            data["longitude"] = point.longitude
        self.post("assets/command/set/", data)

    def set_asset_status(self, asset: SMMAsset, status: SMMMissionAssetStatusValue, notes: str) -> None:
        """
        Set the status of a specific asset in the mission
        """
        data = {
            "value_id": status.id,
            "notes": notes,
        }
        self.post(f"assets/{asset.id}/status/", data)

    def close(self) -> None:
        """
        Close this mission
        """
        self.connection.get(self.__url_component("close/"))

    def assets(self, include: str = "active") -> list[str]:
        """
        Get all the assets in this mission

        Use include="removed" to see get all assets that were ever in the mission
        """
        include_removed = str(include == "removed")
        return self.connection.get_json(self.__url_component(f"assets/?include_removed={include_removed}"))

    def add_waypoint(self, point: SMMPoint, label: str) -> SMMPoi | None:
        """
        Add a way point to this mission
        """
        results = self.post("data/pois/create/", {"lat": point.lat, "lon": point.lng, "label": label})
        if results.status_code == requests.codes["ok"]:
            json_obj = results.json()
            return SMMPoi(self, json_obj["features"][0]["properties"]["pk"])
        return None

    def _populate_points(self, points: list[SMMPoint], label) -> object:
        """
        Add the points to data
        """
        data = {
            "points": len(points),
            "label": label,
        }
        i = 0
        for point in points:
            data[f"point{i}_lat"] = point.lat
            data[f"point{i}_lng"] = point.lng
            i = i + 1
        return data

    def add_line(self, points: list[SMMPoint], label: str) -> SMMLine | None:
        """
        Add a line to this mission
        """
        data = self._populate_points(points, label)
        results = self.post("data/userlines/create/", data)
        if results.status_code == requests.codes["ok"]:
            json_obj = results.json()
            return SMMLine(self, json_obj["features"][0]["properties"]["pk"])
        return None

    def add_polygon(self, points: list[SMMPoint], label: str) -> SMMPolygon | None:
        """
        Add a polygon to this mission
        """
        data = self._populate_points(points, label)
        results = self.post("data/userpolygons/create/", data)
        if results.status_code == requests.codes["ok"]:
            json_obj = results.json()
            return SMMPolygon(self, json_obj["features"][0]["properties"]["pk"])
        return None

    @classmethod
    def get_mission_for_asset(cls, asset: SMMAsset) -> SMMMission | None:
        """
        Get the current mission for the asset
        """
        data = asset.get_asset_data()
        try:
            return SMMMission(asset.connection, data["mission_id"], data["mission_name"])
        except KeyError:
            return None
